#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["HAS_AGENSGRAPH", "HAS_NEO4J"]


try:
    import agensgraph  # noqa: F401
except ImportError:
    HAS_AGENSGRAPH = False
else:
    HAS_AGENSGRAPH = True


try:
    import neo4j  # noqa: F401
except ImportError:
    HAS_NEO4J = False
else:
    HAS_NEO4J = True



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from .session import Session
from ..config import Config

from ..internal.abc import AIOCypherABCMeta

from abc import abstractmethod

from typing import Tuple, Optional, AsyncContextManager


class Driver (object, metaclass=AIOCypherABCMeta):
    """This class encapsulates a driver object for a cypher database. Concrete subclasses specialise in
    different databases.
    """
    def __init__(self,
                 config: Config):
        self._address = config.address
        self._auth: Tuple[str, str] = config.auth

    async def __aenter__(self) -> 'Driver':
        return self

    async def __aexit__(self, e_t, e_v, e_tb) -> Optional[bool]:
        await self.close()
        return False

    @abstractmethod
    def __await__(self):
        ...

    @abstractmethod
    async def close(self) -> None:
        """This coroutine should be awaited when the driver is no longer needed.
        """
        ...

    @abstractmethod
    def session(self, **kwargs) -> AsyncContextManager[Session]:
        """This method is used to create a Session object, which can be used as an
        asynchronous context manager.

        All interactions should be done via a Session object. Accessing the driver
        without one is not supported.
        """
        ...

    @abstractmethod
    def database_type(self) -> str:
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.node import Node as AbstractNode

from typing import Dict, Union


class Node (AbstractNode, Dict[str, Union[str, int]]):
    def __init__(self, data: Dict[str, Union[str, int]]):
        if 'vid' in data:
            self._hash = hash(data['vid'])
        else:
            self._hash = hash(str(data))
        super().__init__({key: value for (key, value) in data.items()})

    def __hash__(self):
        return self._hash



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.result import Result as AbstractResult

from typing import TypeVar, Callable, Awaitable, List, Dict, cast, Any

import neo4j

from .record import Record
from .graph import Graph


R = TypeVar('R')


class Result (AbstractResult[neo4j.Result]):
    """A conceptual wrapper for a neo4j query which will return a neo4j.Result object.

    To execute the query and return the underlying object await this object. But the
    returned neo4j.Result is unlikely to be very useful outside of the context managers
    in which it was created.

    A better way to use this object is to use the 'single()' or 'graph()' methods.
    """
    def __init__(
        self,
        execute: Callable[[Callable[[neo4j.Transaction], R]], Awaitable[R]],
        func: Callable[[neo4j.Transaction], neo4j.Result]
    ):
        self._func = func
        self._execute = execute

    def __await__(self):
        return self._execute(self._func).__await__()

    def single(self) -> Record:
        """Returns an aioneo4j.Record object, which is an Awaitable[neo4j.Record]

        Useful if the wrapped query is expected to return a single result.
        """
        return Record(self._execute, lambda tx: self._func(tx).single())

    def graph(self) -> Graph:
        """Returns an aioneo4j.graph.Graph object which is an Awaitable[neo4j.graph.Graph]

        Useful if the wrapped query is expected to return multiple results.
        """
        return Graph(self._execute, lambda tx: self._func(tx).graph())

    async def data(self) -> List[Dict[str, Any]]:
        """Deserialises the data in this result as a list of dictionaries and returns it.
        """
        return cast(List[Dict[str, Any]], await self._execute(lambda tx: self._func(tx).data()))



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from abc import ABCMeta as ABCMeta

from typing import Type, List, Tuple, Callable, Optional

from ..config import Config

__all__ = [
    "AIOCypherABCMeta"
]


class AIOCypherABCMeta (ABCMeta):

    __concrete_lookup__: List[Tuple[Callable[[Config], bool], Type]] = []

    def __new__(mcls, name, bases, namespace, **kwargs):
        cls = super().__new__(mcls, name, bases, namespace, **kwargs)

        return cls

    def register_concrete(cls, address_parser: Callable[[Config], bool]):
        def __decorator(subclass: Type) -> Type:
            subclass = cls.register(subclass)
            cls.__concrete_lookup__.append((address_parser, subclass))

            return subclass
        return __decorator

    def __call__(cls, *args, **kwargs) -> object:
        if 'config' in kwargs:
            config = kwargs.pop('address')
        elif len(args) > 0:
            config = args[0]
            args = args[1:]
        else:
            raise TypeError("Missing 1 required positional argument (config)")

        concrete_subclass: Optional[Type] = None
        for (address_parser, subclass) in cls.__concrete_lookup__:
            if address_parser(config):
                concrete_subclass = subclass

        if concrete_subclass is None:
            raise ValueError(f"Could not find any registered concrete subclass of {cls!r} for address "
                             f"{config.address}")

        return ABCMeta.__call__(concrete_subclass, config, *args, **kwargs)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.transaction import Transaction as AbstractTransaction

from typing import Optional, TypeVar, Callable, Awaitable

import neo4j

from .result import Result

R = TypeVar('R')


class Transaction (AbstractTransaction):
    """A class which encapsulates a neo4j.Transaction object. Should never be created directly,
    but always via an aioneo4j.Session object.

    Should always be used as an Async Context Manager. Using any of its methods outside of
    the context may lead to undefined behaviour.
    """
    def __init__(self, session_execute: Callable[[Callable[[neo4j.Session], R]], Awaitable[R]]):
        self._session_execute = session_execute
        self._sync_transaction: Optional[neo4j.Transaction] = None

    async def __aenter__(self) -> "Transaction":
        self._sync_transaction = await self._session_execute(lambda session: session.begin_transaction().__enter__())
        return self

    async def __aexit__(self, *args):
        if self._sync_transaction is not None:
            return await self._session_execute(lambda session: self._sync_transaction.__exit__(*args))
        else:
            return False

    async def _execute(self, f: Callable[[neo4j.Transaction], R]) -> R:
        """Although not strictly part of the public interface this coroutine
        may occasionally be useful when doing things not yet supported by
        the interface methods.

        When awaited it executes a provided Synchronous callable on the IO thread and
        returns the return value. If the callable raises an exception then this is
        propogated out of this await.

        :param f: A callable which takes a neo4j.Transaction object as a parameter.
        :returns: The return value of f
        :raises: Any exception raised by f
        """
        return await self._session_execute(lambda session: f(self._sync_transaction))

    def run(self, *args, **kwargs) -> Result:
        """This is the main routine for this class.

        It has an interface identical to neo4j.Transaction.run.

        The actual query will not be executed until such a time as the return value (or some derivative thereof)
        is awaited.

        :returns: an aioneo4j.Result object, which is a subclass of Awaitable[neo4j.Result]
        """
        return Result(self._execute, lambda tx: tx.run(*args, **kwargs))



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import TypeVar, Optional, Callable
import logging
import functools
import time


T = TypeVar('T')


def retry_on_exception(
    exception=Exception,
    timeout=20.0,
    logger: Optional[logging.Logger] = None,
    msg="Retrying ..."
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """This method returns a decorator which causes the decorated function to
    retry repeatedly for a specified timeout if an exception is raised. If the
    timeout is exceded then any further exception is raised as normal.
    :param exception: An exception class to restrict to, any other exceptions will be raised as normal
    :param timeout: the timeout in seconds
    :param logger: a logger to use to log messages
    :param msg: A message to log each time an exception is suppressed and the call is retried
    :returns: A decorator
    """
    def __decorator(f: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(f)
        async def __inner(*args, **kwargs) -> T:
            start = time.clock_gettime(time.CLOCK_MONOTONIC)
            while True:
                try:
                    return await f(*args, **kwargs)
                except exception as e:
                    if time.clock_gettime(time.CLOCK_MONOTONIC) - start > timeout:
                        raise Exception("Timeout exceeded whilst retrying") from e
                    else:
                        if logger is not None:
                            logger.warning(msg)
                        time.sleep(1)
                else:
                    break
        return __inner
    return __decorator



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A limited asyncio wrapper for the neo4j driver

This is not intended to be a full driver wrapper, but just to implement the things
we immediately need in a relatively simple fashion, replicating the interface of the wrapped
library where possible.
"""

from .driver import Driver
from .session import Session
from .transaction import Transaction
from .result import Result
from .record import Record

__all__ = [
    'Driver',
    'Session',
    'Transaction',
    'Result',
    'Record'
]



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.driver import Driver as AbstractDriver

from typing import Tuple, AsyncContextManager, Optional

from urllib.parse import urlparse
import re

import aiopg
import agensgraph as ag  # noqa: F401

from async_exit_stack import AsyncExitStack

from .session import Session
from ..config import Config


def is_agensgraph_address(config: Config) -> bool:
    address = config.address
    return bool(re.match(r'^postgresql(\+.+)?$', urlparse(address).scheme))


def _construct_dsn(address: str, auth: Tuple[str, str]) -> str:
    addr = urlparse(address)
    hostname = addr.hostname or '127.0.0.1'
    port = addr.port or '5432'
    return f"host={hostname} user={auth[0]} password={auth[1]} port={port}"


@AbstractDriver.register_concrete(is_agensgraph_address)
class Driver (AbstractDriver):
    """This class encapsulates an agensgraph db driver
    """
    def __init__(self,
                 config: Config):
        super().__init__(config)
        self._exit_stack = AsyncExitStack()
        self._pool: Optional[aiopg.Pool] = None

        self.dsn = _construct_dsn(self._address, self._auth)

    def database_type(self) -> str:
        return "agensgraph"

    async def __aenter__(self) -> "Driver":
        await self
        return self

    async def __aexit__(self, e_t, e_v, e_tb) -> Optional[bool]:
        return await self._exit_stack.__aexit__(e_t, e_v, e_tb)

    def __await__(self):
        async def __inner():
            if self._pool is None:
                self._pool = await self._exit_stack.enter_async_context(aiopg.create_pool(self.dsn))
            return self._pool
        return __inner().__await__()

    def session(self, **kwargs) -> AsyncContextManager[Session]:
        """This method is used to create a Session object, which can be used as an
        asynchronous context manager.

        All interactions should be done via a Session object. Accessing the driver
        without one is not supported.
        """
        assert (self._pool is not None)
        return Session(self._pool, **kwargs)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest

from .. import HAS_NEO4J
from .fixtures import clean_session


@unittest.skipUnless(HAS_NEO4J, "Don't test aioneo4j unless neo4j is installed")
class TestDriver(unittest.IsolatedAsyncioTestCase):
    @clean_session
    async def test_basic_query(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example'})")

        async with session.begin_transaction() as tx:
            result = await tx.run("MATCH(n:TestNode) RETURN n.name")
            record = result.single()

        self.assertEqual("example", record.value())

    @clean_session
    async def test_single(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example'})")

        async with session.begin_transaction() as tx:
            record = await tx.run("MATCH(n:TestNode) RETURN n.name").single()

        self.assertEqual("example", record.value())

    @clean_session
    async def test_single_value(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example'})")

        async with session.begin_transaction() as tx:
            name = await tx.run("MATCH(n:TestNode) RETURN n.name").single().value()

        self.assertEqual("example", name)

    @clean_session
    async def test_single_data(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")

        async with session.begin_transaction() as tx:
            n = await tx.run("MATCH(n:TestNode) RETURN n").single().data()

        self.assertEqual(n, {
            'n': {
                'name': 'example',
                'potato': 'desirée'
            }
        })

    @clean_session
    async def test_result_data__one_record(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")

        async with session.begin_transaction() as tx:
            n = await tx.run("MATCH(n:TestNode) RETURN n").data()

        self.assertEqual(n, [{
            'n': {
                'name': 'example',
                'potato': 'desirée'
            }
        }])

    @clean_session
    async def test_result_data__two_records(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")

        async with session.begin_transaction() as tx:
            n = await tx.run("MATCH(n:TestNode) RETURN n").data()

        self.assertCountEqual(n, [
            {
                'n': {
                    'name': 'example',
                    'potato': 'desirée'
                }
            },
            {
                'n': {
                    'name': '模範',
                    'potato': 'じゃが芋'
                }
            }
        ])

    @clean_session
    async def test_result_data__relationship(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")
            await tx.run("""\
MATCH(a:TestNode{name:'example'})
MATCH(b:TestNode{name:'模範'})
CREATE (a)-[:Translation {lang:'ja'}]->(b)""")

        async with session.begin_transaction() as tx:
            data = await tx.run("MATCH (a:TestNode)-[r:Translation]-(b:TestNode) RETURN a,r,b").data()

        # It's a weird thing that this doesn't return you the properties of the relationship, but
        # that's down to the underlying synchronous library's behaviour
        self.assertCountEqual(data, [
            {
                'a': {'name': 'example', 'potato': 'desirée'},
                'r': ({'name': 'example', 'potato': 'desirée'}, 'Translation', {'name': '模範', 'potato': 'じゃが芋'}),
                'b': {'name': '模範', 'potato': 'じゃが芋'}
            },
            {
                'b': {'name': 'example', 'potato': 'desirée'},
                'r': ({'name': 'example', 'potato': 'desirée'}, 'Translation', {'name': '模範', 'potato': 'じゃが芋'}),
                'a': {'name': '模範', 'potato': 'じゃが芋'}
            }
        ])

    @clean_session
    async def test_result_data__two_values(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")

        async with session.begin_transaction() as tx:
            n = await tx.run("MATCH(n:TestNode) RETURN n.name AS name, n.potato AS potato").data()

        self.assertCountEqual(n, [
            {
                'name': 'example',
                'potato': 'desirée'
            },
            {
                'name': '模範',
                'potato': 'じゃが芋'
            }
        ])

    @clean_session
    async def test_graph_nodes(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")
            await tx.run("""\
MATCH(a:TestNode{name:'example'})
MATCH(b:TestNode{name:'模範'})
CREATE (a)-[:Translation {lang:'ja'}]->(b)""")

        async with session.begin_transaction() as tx:
            nodes = await tx.run("MATCH (a:TestNode)-[r:Translation]-(b:TestNode) RETURN a,r,b").graph().nodes

        self.assertEqual(len(nodes), 2)
        self.assertCountEqual(
            [n.labels for n in nodes],
            [{'TestNode'}, {'TestNode'}]
        )
        self.assertCountEqual(
            [n['name'] for n in nodes],
            ['example', '模範']
        )
        self.assertCountEqual(
            [n['potato'] for n in nodes],
            ['desirée', 'じゃが芋']
        )

    @clean_session
    async def test_graph_relationships(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")
            await tx.run("""\
MATCH(a:TestNode{name:'example'})
MATCH(b:TestNode{name:'模範'})
CREATE (a)-[:Translation {lang:'ja'}]->(b)""")

        async with session.begin_transaction() as tx:
            rels = await tx.run("MATCH (a:TestNode)-[r:Translation]-(b:TestNode) RETURN a,r,b").graph().relationships

        self.assertEqual(len(rels), 1)
        rel = next(iter(rels))
        self.assertEqual(rel.type, 'Translation')
        self.assertEqual(rel['lang'], 'ja')
        self.assertEqual(rel.start_node.labels, {'TestNode'})
        self.assertEqual(rel.start_node['name'], 'example')
        self.assertEqual(rel.start_node['potato'], 'desirée')
        self.assertEqual(rel.end_node.labels, {'TestNode'})
        self.assertEqual(rel.end_node['name'], '模範')
        self.assertEqual(rel.end_node['potato'], 'じゃが芋')

    @clean_session
    async def test_graph_nodes_and_relationships(self, session):
        async with session.begin_transaction() as tx:
            await tx.run("CREATE(n:TestNode{name:'example', potato:'desirée'})")
            await tx.run("CREATE(n:TestNode{name:'模範', potato:'じゃが芋'})")
            await tx.run("""\
MATCH(a:TestNode{name:'example'})
MATCH(b:TestNode{name:'模範'})
CREATE (a)-[:Translation {lang:'ja'}]->(b)""")

        async with session.begin_transaction() as tx:
            graph = tx.run("MATCH (a:TestNode)-[r:Translation]-(b:TestNode) RETURN a,r,b").graph()
            nodes = await graph.nodes
            rels = await graph.relationships

        self.assertEqual(len(nodes), 2)
        self.assertCountEqual(
            [n.labels for n in nodes],
            [{'TestNode'}, {'TestNode'}]
        )
        self.assertCountEqual(
            [n['name'] for n in nodes],
            ['example', '模範']
        )
        self.assertCountEqual(
            [n['potato'] for n in nodes],
            ['desirée', 'じゃが芋']
        )

        self.assertEqual(len(rels), 1)
        rel = next(iter(rels))
        self.assertEqual(rel.type, 'Translation')
        self.assertEqual(rel['lang'], 'ja')
        self.assertEqual(rel.start_node.labels, {'TestNode'})
        self.assertEqual(rel.start_node['name'], 'example')
        self.assertEqual(rel.start_node['potato'], 'desirée')
        self.assertEqual(rel.end_node.labels, {'TestNode'})
        self.assertEqual(rel.end_node['name'], '模範')
        self.assertEqual(rel.end_node['potato'], 'じゃが芋')



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.session import Session as AbstractSession

from typing import AsyncContextManager, Optional
from async_exit_stack import AsyncExitStack

import aiopg

from .transaction import Transaction


class Session (AbstractSession):
    """This object encapsulates a session with an agensgraph database. It should not be initialised directly, but
    created via an Driver object.

    It should always be used as an Async Context Manager, and using its methods outside of its
    context may lead to undefined behaviour.
    """
    def __init__(self, pool: aiopg.Pool, **kwargs):
        self._exit_stack = AsyncExitStack()
        self._pool = pool
        self._connection: Optional[aiopg.Connection] = None
        self.__kwargs = kwargs

        self._dbname = self.__kwargs.get('database', "database") or "database"

    async def __aenter__(self) -> 'Session':
        await self._exit_stack.__aenter__()
        await self
        return self

    async def __aexit__(self, *args) -> bool:
        return await self._exit_stack.__aexit__(*args)

    def __await__(self):
        async def __inner():
            if self._connection is None:
                self._connection = await self._exit_stack.enter_async_context(self._pool.acquire())
            return self._connection
        return __inner().__await__()

    def begin_transaction(self) -> AsyncContextManager[Transaction]:
        """This is the main method for interacting with the Session. It
        creates a new Transaction which is then represented by an Async Context Manager
        """
        if self._connection is None:
            raise RuntimeError("Cannot begin a transaction unless the session is open")
        return Transaction(self._connection, self._dbname)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractmethod

from typing import Optional, TypeVar, Awaitable, Dict, Union, Generic

from .node import Node

R = TypeVar('R')


class Record (Generic[R], Awaitable[R], metaclass=ABCMeta):
    """A conceptual wrapper for a query which will return a single record object.

    To execute the query and return the underlying object await this object.

    A better way to use this object is to use the 'data()' coroutine.
    """

    @abstractmethod
    async def data(self) -> Optional[Dict[str, Dict[str, Union[str, int]]]]:
        """Deserialises the data in this result as dictionary and returns it.
        """
        ...

    @abstractmethod
    async def value(self) -> Optional[Node]:
        """Return the Node object represented by the query"""
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.session import Session as AbstractSession

import asyncio
import time
from typing import Optional, TypeVar, Callable, AsyncContextManager, Tuple, Any, cast
from queue import Empty, Full

import neo4j
import janus
from async_generator import asynccontextmanager

from .transaction import Transaction


R = TypeVar('R')


class Session (AbstractSession):
    """This object encapsulates a neo4j.Session object. It should not be initialised directly, but
    created via an aioneo4j.Driver object.

    It should always be used as an Async Context Manager, and using its methods outside of its
    context may lead to undefined behaviour.


    Each Session object maintains its own connection thread for performing IO. A Session object is
    the only supported way to interact with the neo4j database.
    """
    def __init__(self, sync_driver: 'asyncio.Future[neo4j.Driver]', **kwargs):
        self.sync_driver = sync_driver
        self.__thread: Optional['asyncio.Future[None]'] = None
        self.__die = False

        self.__kwargs = kwargs

        if "database" in self.__kwargs and self.__kwargs['database'] is None:
            self.__kwargs['database'] = neo4j.DEFAULT_DATABASE

        self._command_queue: janus.Queue[Callable[[neo4j.Transaction], Any]] = janus.Queue(maxsize=1)
        self._response_queue: janus.Queue[Tuple[Any, Optional[BaseException]]] = janus.Queue(maxsize=1)

        self._transaction: Optional[Transaction] = None

    async def __aenter__(self) -> 'Session':
        self.__die = False
        loop = asyncio.get_running_loop()
        self.__thread = cast(
            'asyncio.Future[None]',
            loop.run_in_executor(None, self._thread_main, await self.sync_driver))
        return self

    async def __aexit__(self, *args) -> bool:
        self.__die = True
        if self.__thread is not None:
            await self.__thread
        return False

    def _thread_main(self, sync_driver: neo4j.Driver) -> None:
        with sync_driver.session(**self.__kwargs) as session:
            while not self.__die:
                try:
                    cmd = self._command_queue.sync_q.get(block=False)
                except Empty:
                    time.sleep(0.01)
                    continue

                try:
                    (r, e) = (cmd(session), None)
                except BaseException as ex:
                    (r, e) = (None, ex)

                while not self.__die:
                    try:
                        self._response_queue.sync_q.put((r, e), block=False)
                    except Full:
                        time.sleep(0.01)
                        continue
                    else:
                        break

    async def _execute(self, f: Callable[[neo4j.Session], R]) -> R:
        """Although not strictly part of the public interface this coroutine
        may occasionally be useful when doing things not yet supported by
        the interface methods.

        When awaited it executes a provided Synchronous callable on the IO thread and
        returns the return value. If the callable raises an exception then this is
        propogated out of this await.

        :param f: A callable which takes a neo4j.Session object as a parameter.
        :returns: The return value of f
        :raises: Any exception raised by f
        """
        await self._command_queue.async_q.put(f)
        (r, e) = await self._response_queue.async_q.get()
        if e is not None:
            raise e
        return r

    def begin_transaction(self) -> AsyncContextManager[Transaction]:
        """This is the main method for interacting with the Session. It
        creates a new Transaction which is then represented by an Async Context Manager
        yielding an aioneo4j.Transaction object.
        """
        @asynccontextmanager
        async def __inner():
            if self._transaction is not None:
                raise neo4j.exceptions.TransactionError()

            self._transaction = Transaction(self._execute)

            async with self._transaction:
                yield self._transaction

            self._transaction = None

        return __inner()



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.result import Result as AbstractResult
from ..interface.exceptions import QueryFailed

from typing import List, Dict, Any

import aiopg

import re

from .record import Record
from .graph import Graph


def convert_cypher_parameters_to_postgres(query: str, params: Dict[str, str]) -> str:
    q = str(query)
    for key in params:
        # Replace entries of the form $name with %(name)s
        # (Ignoring instances of $$)
        q = re.sub(r'(?<!\$)\$' + key + r'(?!\w)', r'%(' + key + ')s', q)

    return q


class Result (AbstractResult[Any]):
    """A conceptual wrapper for a query which will return a result of some sort.

    To execute the query and return the underlying object await this object. But the
    returned object is unlikely to be very useful outside of the context managers
    in which it was created.

    A better way to use this object is to use the 'data()', single()' or 'graph()' methods.
    """
    def __init__(self, cursor: aiopg.Cursor, *args, **kwargs):
        self._cursor = cursor
        self._args = args
        self._kwargs = kwargs

    def __await__(self):
        async def __inner():
            query = convert_cypher_parameters_to_postgres(self._args[0], self._kwargs)
            try:
                await self._cursor.execute(query, parameters=self._kwargs)
            except Exception as e:
                raise QueryFailed(query, self._kwargs) from e
            else:
                return self._cursor
        return __inner().__await__()

    def single(self) -> Record:
        """Returns an Record object

        Useful if the wrapped query is expected to return a single result.
        """
        async def __inner():
            await self
            result = await self._cursor.fetchone()
            return result[0] if result else None
        return Record(__inner())

    def graph(self) -> Graph:
        """Returns an Graph object

        Useful if the wrapped query is expected to return multiple results.
        """
        async def __inner():
            await self
            result = await self._cursor.fetchall()
            return result
        return Graph(__inner())

    async def data(self) -> List[Dict[str, Any]]:
        """Deserialises the data in this result as a list of dictionaries and returns it.
        """
        return []



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractmethod

from typing import TypeVar, Awaitable, List, Dict, Any, Generic

from .record import Record
from .graph import Graph


R = TypeVar('R')


class Result (Generic[R], Awaitable[R], metaclass=ABCMeta):
    """A conceptual wrapper for a query which will return a result object of some kind.

    To execute the query and return the underlying object await this object.

    It also has other methods which return awaitable objects.
    """

    @abstractmethod
    def single(self) -> Record:
        """Returns a Record object

        Useful if the wrapped query is expected to return a single result.
        """
        ...

    @abstractmethod
    async def data(self) -> List[Dict[str, Any]]:
        """Deserialises the data in this result as a list of dictionaries and returns it.
        """
        ...

    @abstractmethod
    def graph(self) -> Graph:
        """Returns an Graph object which

        Useful if the wrapped query is expected to return multiple results.
        """
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Dict, Any, Set

from .interface.result import Result
from .interface.record import Record
from .interface.graph import Graph
from .interface.node import Node
from .interface.relationship import Relationship


class EmptyGraph (Graph[None]):
    @property
    async def nodes(self) -> Set[Node]:
        return set()

    @property
    async def relationships(self) -> Set[Relationship]:
        return set()

    def __await__(self):
        async def __inner():
            return None
        return __inner().__await__()


class EmptyResult(Result[None]):
    def single(self) -> Record:
        """Returns a Record object

        Useful if the wrapped query is expected to return a single result.
        """
        raise RuntimeError("This result is empty")

    async def data(self) -> List[Dict[str, Any]]:
        """Deserialises the data in this result as a list of dictionaries and returns it.
        """
        return []

    def graph(self) -> EmptyGraph:
        """Returns an Graph object which

        Useful if the wrapped query is expected to return multiple results.
        """
        return EmptyGraph()

    def __await__(self):
        async def __inner():
            return None
        return __inner().__await__()



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.transaction import Transaction as AbstractTransaction

from typing import Optional

import aiopg
from async_exit_stack import AsyncExitStack

from .result import Result


class Transaction (AbstractTransaction):
    """A class which encapsulates a Transaction. Should never be created directly,
    but always via an Session object.

    Should always be used as an Async Context Manager. Using any of its methods outside of
    the context may lead to undefined behaviour.
    """
    def __init__(self, connection: aiopg.Connection, database: str):
        self._exit_stack = AsyncExitStack()
        self._connection = connection
        self._dbname = database
        self._cursor: Optional[aiopg.Cursor] = None

    def __await__(self):
        async def __inner():
            if self._cursor is None:
                self._cursor = await self._exit_stack.enter_async_context(self._connection.cursor())
                await self._cursor.execute(f"CREATE GRAPH IF NOT EXISTS {self._dbname}")
                await self._cursor.execute(f"SET graph_path = {self._dbname}")
                await self._cursor.execute("BEGIN")
                await self._cursor.execute("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE")
            return self._cursor
        return __inner().__await__()

    async def __aenter__(self) -> "Transaction":
        await self._exit_stack.__aenter__()
        await self
        return self

    async def __aexit__(self, exc_t, exc_v, exc_tb):
        if self._cursor is not None:
            if exc_v is None:
                try:
                    await self._cursor.execute("COMMIT")
                except Exception:
                    pass
            else:
                try:
                    await self._cursor.execute("ROLLBACK")
                except Exception:
                    pass
        return await self._exit_stack.__aexit__(exc_t, exc_v, exc_tb)

    def run(self, *args, **kwargs) -> Result:
        """This is the main routine for this class.

        :returns: a Result object
        """
        return Result(self._cursor, *args, **kwargs)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from async_generator import asynccontextmanager
import os
import asyncio
from functools import wraps


from aiocypher.aioneo4j.driver import Driver
from aiocypher import Config


def clean_session(f=None):
    """This helper gives you a Session on the database, but does some extra work before and after to
    make sure everything is deleted.

    It can be used as an async context manager or a decorator
    """
    @asynccontextmanager
    async def __clean_session():
        HOST = os.environ.get('NEO4J_HOST', 'localhost')
        PORT = os.environ.get('NEO4J_PORT', 17687)
        USER = 'neo4j'
        PASSWORD = 'test'

        DB = "neo4j"

        config = Config(
            address=f'bolt://{HOST}:{PORT}',
            username=USER,
            password=PASSWORD
        )
        async with Driver(config) as driver:
            async with driver.session(database=DB) as session:
                # Delete content in the database
                async with session.begin_transaction() as tx:
                    await tx.run("MATCH (n)\nDETACH DELETE n")

                # Delete any schemata in the database
                async with session.begin_transaction() as tx:
                    await tx.run("CALL apoc.schema.assert({}, {})")

            async with driver.session(database=DB) as session:
                yield session

            async with driver.session(database=DB) as session:
                # Delete content in the database
                async with session.begin_transaction() as tx:
                    await tx.run("MATCH (n)\nDETACH DELETE n")

    if f is None:
        return __clean_session()
    else:
        @wraps(f)
        async def __inner(*args, **kwargs):
            async with __clean_session() as session:
                r = f(*args, session, **kwargs)
                if asyncio.iscoroutinefunction(f):
                    return await r
                else:
                    return r

        return __inner



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractproperty

from typing import TypeVar, Awaitable, Generic, Collection

from .relationship import Relationship
from .node import Node


R = TypeVar('R')


class Graph (Generic[R], Awaitable[R], metaclass=ABCMeta):
    """A conceptual wrapper for a query which will return a graph of nodes.

    To execute the query and return the underlying object await this object.

    A better way to use this object is to use the 'nodes' or 'relationships' coroutine properties.
    """
    @abstractproperty
    async def nodes(self) -> Collection[Node]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.

        When awaited this property will execute the query and return you a Set[neo4j.graph.Node]
        containing all of the nodes which the query matched.
        """
        ...

    @abstractproperty
    async def relationships(self) -> Collection[Relationship]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.

        When awaited this property will execute the query and return you a Set[neo4j.graph.Relationship]
        containing all of the relationships which the query matched.
        """
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.graph import Graph as AbstractGraph

from typing import Set, Any, Awaitable, List, Dict, Optional

from .node import Node

Relationship = Any


class Graph (AbstractGraph[Optional[List[Dict[str, str]]]]):
    """A conceptual wrapper for a query which will return a Graph.
    """
    def __init__(
        self,
        coroutine: Awaitable[List[Dict[str, str]]]
    ):
        self._coroutine: Optional[Awaitable[List[Dict[str, str]]]] = coroutine
        self._result: Optional[List[Dict[str, str]]] = None

    def __await__(self):
        async def __inner():
            if self._result is None and self._coroutine is not None:
                c = self._coroutine
                self._coroutine = None
                self._result = await c
            return self._result
        return __inner().__await__()

    @property
    async def nodes(self) -> Set[Node]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.
        """
        data = await self
        return {Node(n) for (n,) in data if 'vid' in n}

    @property
    async def relationships(self) -> Set[Relationship]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.
        """
        return set()



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractmethod
from typing import AsyncContextManager

from .transaction import Transaction


class Session (object, metaclass=ABCMeta):
    """This object encapsulates a Session object. It should not be initialised directly, but
    created via a Driver object.

    It should always be used as an Async Context Manager, and using its methods outside of its
    context may lead to undefined behaviour.

    A Session object is the only supported way to interact with the cypher database.
    """
    async def __aenter__(self) -> 'Session':
        return self

    async def __aexit__(self, *args) -> bool:
        return False

    @abstractmethod
    def begin_transaction(self) -> AsyncContextManager[Transaction]:
        """This is the main method for interacting with the Session. It
        creates a new Transaction which is then represented by an Async Context Manager
        """
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.driver import Driver as AbstractDriver
from .session import Session
from ..config import Config

import asyncio
import time
from typing import Optional, Callable, TypeVar, AsyncContextManager
from functools import wraps

import neo4j
from neo4j.exceptions import ServiceUnavailable
from async_generator import asynccontextmanager

from urllib.parse import urlparse
import re

T = TypeVar('T')


def _wrapped_in_executor(f: Callable[..., T]) -> Callable[..., 'asyncio.Future[T]']:
    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_running_loop()
        return loop.run_in_executor(
            None,
            lambda: f(*args, **kwargs))

    return wrapper


def is_neo4j_address(config: Config) -> bool:
    address = config.address
    return any(re.match(regex, urlparse(address).scheme) for regex in (r'^neo4j(\+.+)?$', r'^bolt(\+.+)?$'))


@AbstractDriver.register_concrete(is_neo4j_address)
class Driver (AbstractDriver):
    """This class encapsulates the neo4j.Driver object, it's interface is similar but
    limited to what has been implemented yet.
    """
    def __init__(self,
                 config: Config):
        super().__init__(config)
        self._sync_driver: Optional['asyncio.Future[neo4j.Driver]'] = None

    def database_type(self) -> str:
        return "neo4j"

    @property
    def sync_driver(self) -> 'asyncio.Future[neo4j.Driver]':
        """This property is an asyncio Future wrapping the underlying neo4j.Driver object.
        """
        @_wrapped_in_executor
        def __create_sync_driver(address, auth, max_retries=20):
            backoff = 0
            while True:
                try:
                    return neo4j.GraphDatabase.driver(address, auth=auth)
                except (ServiceUnavailable, ConnectionRefusedError) as e:
                    backoff += 1
                    if backoff > max_retries:
                        raise TimeoutError from e
                    time.sleep(backoff)

        if self._sync_driver is None:
            self._sync_driver = __create_sync_driver(
                self._address,
                auth=self._auth)

        return self._sync_driver

    def __await__(self):
        return self.sync_driver.__await__()

    async def close(self) -> None:
        """This coroutine should be awaited when the driver is no longer needed.
        """
        if self._sync_driver is not None:
            sync_driver = await self._sync_driver
            await _wrapped_in_executor(sync_driver.close)()
            self._sync_driver = None

    def session(self, **kwargs) -> AsyncContextManager[Session]:
        """This method is used to create a aioneo4j.Session object, which can be used as an
        asynchronous context manager.

        All interactions should be done via a Session object. Accessing the driver
        without one is not supported.
        """
        @asynccontextmanager
        async def __inner(sync_driver: 'asyncio.Future[neo4j.Driver]'):
            session = Session(sync_driver, **kwargs)
            async with session:
                yield session

        return __inner(self.sync_driver)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from abc import ABCMeta, abstractmethod

from .result import Result


class Transaction (object, metaclass=ABCMeta):
    """A class which encapsulates a Transaction. Should never be created directly,
    but always via a Session object.

    Should always be used as an Async Context Manager. Using any of its methods outside of
    the context may lead to undefined behaviour.
    """
    async def __aenter__(self) -> "Transaction":
        return self

    async def __aexit__(self, *args):
        return False

    @abstractmethod
    def run(self, *args, **kwargs) -> Result:
        """This is the main routine for this class.

        The actual query will not be executed until such a time as the return value (or some derivative thereof)
        is awaited.

        :returns: an Result object
        """
        ...



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.record import Record as AbstractRecord

from typing import Optional, TypeVar, Callable, Awaitable, Dict, cast, Union

import neo4j


R = TypeVar('R')


class Record (AbstractRecord[neo4j.Record]):
    """A conceptual wrapper for a neo4j query which will return a neo4j.Record object.

    To execute the query and return the underlying object await this object. But the
    returned neo4j.Record is unlikely to be very useful outside of the context managers
    in which it was created.

    A better way to use this object is to use the 'value()' coroutine.
    """
    def __init__(
        self,
        execute: Callable[[Callable[[neo4j.Transaction], neo4j.Record]], Awaitable[neo4j.Record]],
        func: Callable[[neo4j.Transaction], neo4j.Record]
    ):
        self._func = func
        self._execute = execute

    def __await__(self):
        return self._execute(self._func).__await__()

    async def value(self) -> Optional[neo4j.graph.Node]:
        """When awaited this coroutine executes the query and returns the underlying neo4j.graph.Node object
        which is contained inside this Record.

        This object is likely to be useful and what you actually want.

        :returns: A neo4j.graph.Node
        """
        def __inner(tx) -> Optional[neo4j.graph.Node]:
            record = self._func(tx)
            if record is not None:
                return record.value()
            else:
                return None

        return await self._execute(__inner)

    async def data(self) -> Optional[Dict[str, Dict[str, Union[str, int]]]]:
        """Deserialises the data in this result as dictionary and returns it.
        """
        def __inner(tx) -> Optional[Dict[str, Union[str, int]]]:
            record = self._func(tx)
            if record is not None:
                return record.data()
            else:
                return None

        return cast(Optional[Dict[str, Dict[str, Union[str, int]]]], await self._execute(__inner))



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""A limited asyncio wrapper for the agensgraph driver

This is not intended to be a full driver wrapper, but just to implement the things
we immediately need in a relatively simple fashion, replicating the interface of
aioneo4j where possible.
"""

from typing import List

__all__: List[str] = []

try:
    from .driver import Driver  # noqa: F401
except ImportError:
    pass
else:
    __all__.append('Driver')



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import unittest
import os

import psycopg2

from ..retry_on_exception import retry_on_exception
from .. import HAS_AGENSGRAPH

if HAS_AGENSGRAPH:
    from aiocypher.aioagensgraph.driver import Driver
    from aiocypher import Config
    import aiopg

    HOST = os.environ.get('AGENSGRAPH_HOST', 'localhost')
    PORT = os.environ.get('AGENSGRAPH_PORT', 15432)
    USER = 'neo4j'
    PASSWORD = 'test'

    DB = "testdb"

    def make_driver():
        config = Config(
            address=f'postgresql://{HOST}:{PORT}',
            username=USER,
            password=PASSWORD)
        return Driver(config)

    @retry_on_exception(exception=psycopg2.OperationalError)
    async def _check_db_connection():
        async with aiopg.connect(f"host={HOST} user={USER} password={PASSWORD} port={PORT}"):
            # Database is connected, tests can continue
            pass

    @unittest.skipUnless(HAS_AGENSGRAPH, "Don't test aioagensgraph unless agensgraph is installed")
    class TestDriver(unittest.IsolatedAsyncioTestCase):
        async def setUp(self):
            async with aiopg.connect(f"host={HOST} user={USER} password={PASSWORD} port={PORT}") as conn:
                async with conn.cursor() as cur:
                    await cur.execute(f"DROP GRAPH IF EXISTS {DB} CASCADE")
                    await cur.execute(f"CREATE GRAPH IF NOT EXISTS {DB}")

        async def tearDown(self):
            async with aiopg.connect(f"host={HOST} user={USER} password={PASSWORD} port={PORT}") as conn:
                async with conn.cursor() as cur:
                    await cur.execute(f"DROP GRAPH IF EXISTS {DB} CASCADE")

        async def test_driver_manual_access_to_underlying_connection_pool(self):
            async with make_driver() as UUT:
                pool = await UUT
                async with pool.acquire() as conn:
                    async with conn.cursor() as cur:
                        await cur.execute(f"SET graph_path = {DB}")
                        await cur.execute("CREATE (:v {name: 'AgensGraph'})")
                        await cur.execute("MATCH (n) RETURN n")
                        v = (await cur.fetchone())[0]
            self.assertEqual(v, {'type': 'vertex', 'label': 'v', 'vid': '3.1', 'name': 'AgensGraph'})

        async def test_driver_session_manual_access_to_underlying_connection(self):
            async with make_driver() as UUT:
                async with UUT.session(database=DB) as session:
                    conn = await session

                    async with conn.cursor() as cur:
                        await cur.execute(f"SET graph_path = {DB}")
                        await cur.execute("CREATE (:v {name: 'AgensGraph'})")
                        await cur.execute("MATCH (n) RETURN n")
                        v = (await cur.fetchone())[0]
            self.assertEqual(v, {'type': 'vertex', 'label': 'v', 'vid': '3.1', 'name': 'AgensGraph'})

        async def test_driver_session_transaction_manual_access_to_underlying_cursor(self):
            async with make_driver() as UUT:
                async with UUT.session(database=DB) as session:
                    async with session.begin_transaction() as tx:
                        cur = await tx

                        await cur.execute("CREATE (:v {name: 'AgensGraph'})")
                        await cur.execute("MATCH (n) RETURN n")
                        v = (await cur.fetchone())[0]
            self.assertEqual(v, {'type': 'vertex', 'label': 'v', 'vid': '3.1', 'name': 'AgensGraph'})

        async def test_driver_session_transaction_run_result_manual_access_to_underlying_cursor(self):
            async with make_driver() as UUT:
                async with UUT.session(database=DB) as session:
                    async with session.begin_transaction() as tx:
                        await tx.run("CREATE (:v {name: 'AgensGraph'})")
                        cur = await tx.run("MATCH (n) RETURN n")
                        v = (await cur.fetchone())[0]
            self.assertEqual(v, {'type': 'vertex', 'label': 'v', 'vid': '3.1', 'name': 'AgensGraph'})

        async def test_driver_session_transaction_run_result_single_manual_access_to_underlying_data(self):
            async with make_driver() as UUT:
                async with UUT.session(database=DB) as session:
                    async with session.begin_transaction() as tx:
                        await tx.run("CREATE (:v {name: 'AgensGraph'})")
                        v = await tx.run("MATCH (n) RETURN n").single()
            self.assertEqual(v, {'type': 'vertex', 'label': 'v', 'vid': '3.1', 'name': 'AgensGraph'})

        async def test_driver_session_transaction_run_result_single_value(self):
            async with make_driver() as UUT:
                async with UUT.session(database=DB) as session:
                    async with session.begin_transaction() as tx:
                        await tx.run("CREATE (:v {name: 'AgensGraph'})")
                        node = await tx.run("MATCH (n) RETURN n").single().value()
            self.assertIn('name', node)
            self.assertEqual('AgensGraph', node['name'])



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.record import Record as AbstractRecord

from typing import Optional, Awaitable, Dict, Union

from .node import Node


class Record (AbstractRecord[Optional[Dict[str, str]]]):
    """A conceptual wrapper for a query which will return a single record.

    To execute the query and return the underlying object await this object or the value() coroutine.

    Both produce slightly different flavours of dictionary representing the data.
    """
    def __init__(
        self,
        coroutine: Awaitable[Optional[Dict[str, str]]]
    ):
        self._coroutine: Optional[Awaitable[Optional[Dict[str, str]]]] = coroutine
        self._result: Optional[Dict[str, str]] = None

    def __await__(self):
        async def __inner():
            if self._result is None and self._coroutine is not None:
                c = self._coroutine
                self._coroutine = None
                self._result = await c

            return self._result
        return __inner().__await__()

    async def value(self) -> Optional[Node]:
        """When awaited this coroutine executes the query and returns a Node object
        which is contained inside this Record.

        This object is likely to be useful and what you actually want.

        :returns: A Node
        """
        result = await self
        if result is None:
            return None
        else:
            return Node(result)

    async def data(self) -> Optional[Dict[str, Dict[str, Union[str, int]]]]:
        """Deserialises the data in this result as dictionary and returns it.
        """
        result = await self

        if result is None or 'label' not in result:
            return None

        label = result['label']

        return {label: Node(result)}



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from setuptools import setup

# Basic metadata
name = 'aiocypher'
description = 'AsyncIO wrapper around the neo4j driver'
url = 'https://github.com/bbc/rd-cloudfit-python-aiocypher'
author = 'BBC R&D'
author_email = 'cloudfit-opensource@rd.bbc.co.uk'
license = 'License :: OSI Approved :: Apache Software License'
long_description = description


# Execute version file to set version variable
try:
    with open(("{}/_version.py".format(name)), "r") as fp:
        exec(fp.read())
except IOError:
    # Version file doesn't exist, fake it for now
    __version__ = "0.0.0"

package_names = [
    'aiocypher',
    'aiocypher.aioagensgraph',
    'aiocypher.aioneo4j',
    'aiocypher.interface',
    'aiocypher.internal'
]
packages = {
    pkg: pkg.replace('.', '/') for pkg in package_names
}

packages_required = [
    "async_generator",
    "neo4j < 4.3.0",
    "janus",
]

extras_require = {
    "agensgraph": [
        "agensgraph",
        "aiopg",
        "async_exit_stack"
    ],
    "neo4j": []
}

setup(name=name,
      python_requires='>=3.10',
      version=__version__,
      description=description,
      url=url,
      author=author,
      author_email=author_email,
      license=license,
      packages=package_names,
      package_dir=packages,
      package_data={
          package_name: ["py.typed"] for package_name in package_names
      },
      install_requires=packages_required,
      extras_require=extras_require,
      scripts=[],
      data_files=[],
      long_description=long_description)



#
#
# Copyright 2020-21 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from ..interface.graph import Graph as AbstractGraph

from typing import TypeVar, Callable, Awaitable, Set

import neo4j


R = TypeVar('R')


class Graph (AbstractGraph[neo4j.graph.Graph]):
    """A conceptual wrapper for a neo4j query which will return a neo4j.graph.Graph object.

    To execute the query and return the underlying object await this object. But the
    returned neo4j.graph.Graph is unlikely to be very useful outside of the context managers
    in which it was created.

    A better way to use this object is to use the 'nodes' coroutine property.
    """
    def __init__(
        self,
        execute: Callable[[Callable[[neo4j.Transaction], neo4j.graph.Graph]], Awaitable[neo4j.graph.Graph]],
        func: Callable[[neo4j.Transaction], neo4j.graph.Graph]
    ):
        self._func = func
        self._execute = execute

    def __await__(self):
        return self._execute(self._func).__await__()

    @property
    async def nodes(self) -> Set[neo4j.graph.Node]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.

        When awaited this property will execute the query and return you a Set[neo4j.graph.Node]
        containing all of the nodes which the query matched.
        """
        return await self._execute(lambda tx: set(self._func(tx).nodes))

    @property
    async def relationships(self) -> Set[neo4j.graph.Relationship]:
        """This property is a Coroutine, which is weird, but better matches the neo4j interface.

        When awaited this property will execute the query and return you a Set[neo4j.graph.Relationship]
        containing all of the relationships which the query matched.
        """
        return await self._execute(lambda tx: set(self._func(tx).relationships))
